# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from time import time
from typing import Any, Callable, Dict, Iterable, List, Optional

from sentry_sdk import capture_exception, push_scope

try:
    from systemd.daemon import notify
except ImportError:
    notify = None

from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)

from swh.core.statsd import statsd
from swh.journal.fixer import fix_objects
from swh.model.hashutil import hash_to_hex

from swh.model.model import (
    BaseContent,
    BaseModel,
    Content,
    Directory,
    Origin,
    OriginVisit,
    Revision,
    SHA1_SIZE,
    SkippedContent,
    Snapshot,
    Release,
)
from swh.objstorage.objstorage import (
    ID_HASH_ALGO,
    ObjNotFoundError,
    ObjStorage,
)
from swh.storage.exc import HashCollision

logger = logging.getLogger(__name__)

GRAPH_OPERATIONS_METRIC = "swh_graph_replayer_operations_total"
GRAPH_DURATION_METRIC = "swh_graph_replayer_duration_seconds"
CONTENT_OPERATIONS_METRIC = "swh_content_replayer_operations_total"
CONTENT_RETRY_METRIC = "swh_content_replayer_retries_total"
CONTENT_BYTES_METRIC = "swh_content_replayer_bytes"
CONTENT_DURATION_METRIC = "swh_content_replayer_duration_seconds"


object_converter_fn: Dict[str, Callable[[Dict], BaseModel]] = {
    "origin": Origin.from_dict,
    "origin_visit": OriginVisit.from_dict,
    "snapshot": Snapshot.from_dict,
    "revision": Revision.from_dict,
    "release": Release.from_dict,
    "directory": Directory.from_dict,
    "content": Content.from_dict,
    "skipped_content": SkippedContent.from_dict,
}


def process_replay_objects(all_objects, *, storage):
    for (object_type, objects) in all_objects.items():
        logger.debug("Inserting %s %s objects", len(objects), object_type)
        with statsd.timed(GRAPH_DURATION_METRIC, tags={"object_type": object_type}):
            _insert_objects(object_type, objects, storage)
        statsd.increment(
            GRAPH_OPERATIONS_METRIC, len(objects), tags={"object_type": object_type}
        )
    if notify:
        notify("WATCHDOG=1")


def collision_aware_content_add(
    content_add_fn: Callable[[Iterable[Any]], None], contents: List[BaseContent]
) -> None:
    """Add contents to storage. If a hash collision is detected, an error is
       logged. Then this adds the other non colliding contents to the storage.

    Args:
        content_add_fn: Storage content callable
        contents: List of contents or skipped contents to add to storage

    """
    if not contents:
        return
    colliding_content_hashes: List[Dict[str, Any]] = []
    while True:
        try:
            content_add_fn(contents)
        except HashCollision as e:
            colliding_content_hashes.append(
                {
                    "algo": e.algo,
                    "hash": e.hash_id,  # hex hash id
                    "objects": e.colliding_contents,  # hex hashes
                }
            )
            colliding_hashes = e.colliding_content_hashes()
            # Drop the colliding contents from the transaction
            contents = [c for c in contents if c.hashes() not in colliding_hashes]
        else:
            # Successfully added contents, we are done
            break
    if colliding_content_hashes:
        for collision in colliding_content_hashes:
            logger.error("Collision detected: %(collision)s", {"collision": collision})


def _insert_objects(object_type: str, objects: List[Dict], storage) -> None:
    """Insert objects of type object_type in the storage.

    """
    objects = fix_objects(object_type, objects)

    if object_type == "content":
        contents: List[BaseContent] = []
        skipped_contents: List[BaseContent] = []
        for content in objects:
            c = BaseContent.from_dict(content)
            if isinstance(c, SkippedContent):
                skipped_contents.append(c)
            else:
                contents.append(c)

        collision_aware_content_add(storage.skipped_content_add, skipped_contents)
        collision_aware_content_add(storage.content_add_metadata, contents)
    elif object_type == "origin_visit":
        visits: List[OriginVisit] = []
        origins: List[Origin] = []
        for obj in objects:
            visit = OriginVisit.from_dict(obj)
            visits.append(visit)
            origins.append(Origin(url=visit.origin))
        storage.origin_add(origins)
        storage.origin_visit_upsert(visits)
    elif object_type in ("directory", "revision", "release", "snapshot", "origin"):
        method = getattr(storage, object_type + "_add")
        method(object_converter_fn[object_type](o) for o in objects)
    else:
        logger.warning("Received a series of %s, this should not happen", object_type)


def is_hash_in_bytearray(hash_, array, nb_hashes, hash_size=SHA1_SIZE):
    """
    Checks if the given hash is in the provided `array`. The array must be
    a *sorted* list of sha1 hashes, and contain `nb_hashes` hashes
    (so its size must by `nb_hashes*hash_size` bytes).

    Args:
        hash_ (bytes): the hash to look for
        array (bytes): a sorted concatenated array of hashes (may be of
            any type supporting slice indexing, eg. :class:`mmap.mmap`)
        nb_hashes (int): number of hashes in the array
        hash_size (int): size of a hash (defaults to 20, for SHA1)

    Example:

    >>> import os
    >>> hash1 = os.urandom(20)
    >>> hash2 = os.urandom(20)
    >>> hash3 = os.urandom(20)
    >>> array = b''.join(sorted([hash1, hash2]))
    >>> is_hash_in_bytearray(hash1, array, 2)
    True
    >>> is_hash_in_bytearray(hash2, array, 2)
    True
    >>> is_hash_in_bytearray(hash3, array, 2)
    False
    """
    if len(hash_) != hash_size:
        raise ValueError("hash_ does not match the provided hash_size.")

    def get_hash(position):
        return array[position * hash_size : (position + 1) * hash_size]

    # Regular dichotomy:
    left = 0
    right = nb_hashes
    while left < right - 1:
        middle = int((right + left) / 2)
        pivot = get_hash(middle)
        if pivot == hash_:
            return True
        elif pivot < hash_:
            left = middle
        else:
            right = middle
    return get_hash(left) == hash_


class ReplayError(Exception):
    """An error occurred during the replay of an object"""

    def __init__(self, operation, *, obj_id, exc):
        self.operation = operation
        self.obj_id = hash_to_hex(obj_id)
        self.exc = exc

    def __str__(self):
        return "ReplayError(doing %s, %s, %s)" % (self.operation, self.obj_id, self.exc)


def log_replay_retry(retry_obj, sleep, last_result):
    """Log a retry of the content replayer"""
    exc = last_result.exception()
    logger.debug(
        "Retry operation %(operation)s on %(obj_id)s: %(exc)s",
        {"operation": exc.operation, "obj_id": exc.obj_id, "exc": str(exc.exc)},
    )

    statsd.increment(
        CONTENT_RETRY_METRIC,
        tags={
            "operation": exc.operation,
            "attempt": str(retry_obj.statistics["attempt_number"]),
        },
    )


def log_replay_error(last_attempt):
    """Log a replay error to sentry"""
    exc = last_attempt.exception()
    with push_scope() as scope:
        scope.set_tag("operation", exc.operation)
        scope.set_extra("obj_id", exc.obj_id)
        capture_exception(exc.exc)

    logger.error(
        "Failed operation %(operation)s on %(obj_id)s after %(retries)s"
        " retries: %(exc)s",
        {
            "obj_id": exc.obj_id,
            "operation": exc.operation,
            "exc": str(exc.exc),
            "retries": last_attempt.attempt_number,
        },
    )

    return None


CONTENT_REPLAY_RETRIES = 3

content_replay_retry = retry(
    retry=retry_if_exception_type(ReplayError),
    stop=stop_after_attempt(CONTENT_REPLAY_RETRIES),
    wait=wait_random_exponential(multiplier=1, max=60),
    before_sleep=log_replay_retry,
    retry_error_callback=log_replay_error,
)


@content_replay_retry
def copy_object(obj_id, src, dst):
    hex_obj_id = hash_to_hex(obj_id)
    obj = ""
    try:
        with statsd.timed(CONTENT_DURATION_METRIC, tags={"request": "get"}):
            obj = src.get(obj_id)
            logger.debug("retrieved %(obj_id)s", {"obj_id": hex_obj_id})

        with statsd.timed(CONTENT_DURATION_METRIC, tags={"request": "put"}):
            dst.add(obj, obj_id=obj_id, check_presence=False)
            logger.debug("copied %(obj_id)s", {"obj_id": hex_obj_id})
        statsd.increment(CONTENT_BYTES_METRIC, len(obj))
    except ObjNotFoundError:
        logger.error(
            "Failed to copy %(obj_id)s: object not found", {"obj_id": hex_obj_id}
        )
        raise
    except Exception as exc:
        raise ReplayError("copy", obj_id=obj_id, exc=exc) from None
    return len(obj)


@content_replay_retry
def obj_in_objstorage(obj_id, dst):
    """Check if an object is already in an objstorage, tenaciously"""
    try:
        return obj_id in dst
    except Exception as exc:
        raise ReplayError("in_dst", obj_id=obj_id, exc=exc) from None


def process_replay_objects_content(
    all_objects: Dict[str, List[dict]],
    *,
    src: ObjStorage,
    dst: ObjStorage,
    exclude_fn: Optional[Callable[[dict], bool]] = None,
    check_dst: bool = True,
):
    """
    Takes a list of records from Kafka (see
    :py:func:`swh.journal.client.JournalClient.process`) and copies them
    from the `src` objstorage to the `dst` objstorage, if:

    * `obj['status']` is `'visible'`
    * `exclude_fn(obj)` is `False` (if `exclude_fn` is provided)
    * `obj['sha1'] not in dst` (if `check_dst` is True)

    Args:
        all_objects: Objects passed by the Kafka client. Most importantly,
            `all_objects['content'][*]['sha1']` is the sha1 hash of each
            content.
        src: An object storage (see :py:func:`swh.objstorage.get_objstorage`)
        dst: An object storage (see :py:func:`swh.objstorage.get_objstorage`)
        exclude_fn: Determines whether an object should be copied.
        check_dst: Determines whether we should check the destination
            objstorage before copying.

    Example:

    >>> from swh.objstorage import get_objstorage
    >>> src = get_objstorage('memory', {})
    >>> dst = get_objstorage('memory', {})
    >>> id1 = src.add(b'foo bar')
    >>> id2 = src.add(b'baz qux')
    >>> kafka_partitions = {
    ...     'content': [
    ...         {
    ...             'sha1': id1,
    ...             'status': 'visible',
    ...         },
    ...         {
    ...             'sha1': id2,
    ...             'status': 'visible',
    ...         },
    ...     ]
    ... }
    >>> process_replay_objects_content(
    ...     kafka_partitions, src=src, dst=dst,
    ...     exclude_fn=lambda obj: obj['sha1'] == id1)
    >>> id1 in dst
    False
    >>> id2 in dst
    True
    """
    vol = []
    nb_skipped = 0
    nb_failures = 0
    t0 = time()

    for (object_type, objects) in all_objects.items():
        if object_type != "content":
            logger.warning(
                "Received a series of %s, this should not happen", object_type
            )
            continue
        for obj in objects:
            obj_id = obj[ID_HASH_ALGO]
            if obj["status"] != "visible":
                nb_skipped += 1
                logger.debug(
                    "skipped %s (status=%s)", hash_to_hex(obj_id), obj["status"]
                )
                statsd.increment(
                    CONTENT_OPERATIONS_METRIC,
                    tags={"decision": "skipped", "status": obj["status"]},
                )
            elif exclude_fn and exclude_fn(obj):
                nb_skipped += 1
                logger.debug("skipped %s (manually excluded)", hash_to_hex(obj_id))
                statsd.increment(
                    CONTENT_OPERATIONS_METRIC, tags={"decision": "excluded"}
                )
            elif check_dst and obj_in_objstorage(obj_id, dst):
                nb_skipped += 1
                logger.debug("skipped %s (in dst)", hash_to_hex(obj_id))
                statsd.increment(CONTENT_OPERATIONS_METRIC, tags={"decision": "in_dst"})
            else:
                try:
                    copied = copy_object(obj_id, src, dst)
                except ObjNotFoundError:
                    nb_skipped += 1
                    statsd.increment(
                        CONTENT_OPERATIONS_METRIC, tags={"decision": "not_in_src"}
                    )
                else:
                    if copied is None:
                        nb_failures += 1
                        statsd.increment(
                            CONTENT_OPERATIONS_METRIC, tags={"decision": "failed"}
                        )
                    else:
                        vol.append(copied)
                        statsd.increment(
                            CONTENT_OPERATIONS_METRIC, tags={"decision": "copied"}
                        )

    dt = time() - t0
    logger.info(
        "processed %s content objects in %.1fsec "
        "(%.1f obj/sec, %.1fMB/sec) - %d failed - %d skipped",
        len(vol),
        dt,
        len(vol) / dt,
        sum(vol) / 1024 / 1024 / dt,
        nb_failures,
        nb_skipped,
    )

    if notify:
        notify("WATCHDOG=1")



# Copyright (C) 2018-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict

from confluent_kafka import Consumer, Producer, KafkaException

import pytest
from subprocess import Popen
from typing import List, Tuple

from swh.storage import get_storage

from swh.journal.serializers import object_key, kafka_to_key, kafka_to_value
from swh.journal.writer.kafka import KafkaJournalWriter, KafkaDeliveryError

from swh.model.model import Directory, DirectoryEntry, Origin, OriginVisit

from .conftest import TEST_OBJECTS, TEST_OBJECT_DICTS


def consume_messages(consumer, kafka_prefix, expected_messages):
    """Consume expected_messages from the consumer;
    Sort them all into a consumed_objects dict"""
    consumed_messages = defaultdict(list)

    fetched_messages = 0
    retries_left = 1000

    while fetched_messages < expected_messages:
        if retries_left == 0:
            raise ValueError("Timed out fetching messages from kafka")

        msg = consumer.poll(timeout=0.01)

        if not msg:
            retries_left -= 1
            continue

        error = msg.error()
        if error is not None:
            if error.fatal():
                raise KafkaException(error)
            retries_left -= 1
            continue

        fetched_messages += 1
        topic = msg.topic()
        assert topic.startswith(kafka_prefix + "."), "Unexpected topic"
        object_type = topic[len(kafka_prefix + ".") :]

        consumed_messages[object_type].append(
            (kafka_to_key(msg.key()), kafka_to_value(msg.value()))
        )

    return consumed_messages


def assert_all_objects_consumed(consumed_messages):
    """Check whether all objects from TEST_OBJECT_DICTS have been consumed"""
    for object_type, known_values in TEST_OBJECT_DICTS.items():
        known_keys = [object_key(object_type, obj) for obj in TEST_OBJECTS[object_type]]

        (received_keys, received_values) = zip(*consumed_messages[object_type])

        if object_type == "origin_visit":
            for value in received_values:
                del value["visit"]
        elif object_type == "content":
            for value in received_values:
                del value["ctime"]

        for key in known_keys:
            assert key in received_keys

        for value in known_values:
            assert value in received_values


def test_kafka_writer(
    kafka_prefix: str, kafka_server: Tuple[Popen, int], consumer: Consumer
):
    kafka_prefix += ".swh.journal.objects"

    writer = KafkaJournalWriter(
        brokers=[f"localhost:{kafka_server[1]}"],
        client_id="kafka_writer",
        prefix=kafka_prefix,
    )

    expected_messages = 0

    for object_type, objects in TEST_OBJECTS.items():
        writer.write_additions(object_type, objects)
        expected_messages += len(objects)

    consumed_messages = consume_messages(consumer, kafka_prefix, expected_messages)
    assert_all_objects_consumed(consumed_messages)


def test_storage_direct_writer(
    kafka_prefix: str, kafka_server: Tuple[Popen, int], consumer: Consumer
):
    kafka_prefix += ".swh.journal.objects"

    writer_config = {
        "cls": "kafka",
        "brokers": ["localhost:%d" % kafka_server[1]],
        "client_id": "kafka_writer",
        "prefix": kafka_prefix,
    }
    storage_config = {
        "cls": "pipeline",
        "steps": [{"cls": "memory", "journal_writer": writer_config},],
    }

    storage = get_storage(**storage_config)

    expected_messages = 0

    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(storage, object_type + "_add")
        if object_type in (
            "content",
            "directory",
            "revision",
            "release",
            "snapshot",
            "origin",
        ):
            method(objects)
            expected_messages += len(objects)
        elif object_type in ("origin_visit",):
            for obj in objects:
                assert isinstance(obj, OriginVisit)
                storage.origin_add_one(Origin(url=obj.origin))
                visit = method(obj.origin, date=obj.date, type=obj.type)
                expected_messages += 1

                obj_d = obj.to_dict()
                for k in ("visit", "origin", "date", "type"):
                    del obj_d[k]
                storage.origin_visit_update(obj.origin, visit.visit, **obj_d)
                expected_messages += 1
        else:
            assert False, object_type

    consumed_messages = consume_messages(consumer, kafka_prefix, expected_messages)
    assert_all_objects_consumed(consumed_messages)


@pytest.fixture(scope="session")
def large_directories() -> List[Directory]:
    dir_sizes = [1 << n for n in range(21)]  # 2**0 = 1 to 2**20 = 1024 * 1024

    dir_entries = [
        DirectoryEntry(
            name=("%09d" % i).encode(),
            type="file",
            perms=0o100644,
            target=b"\x00" * 20,
        )
        for i in range(max(dir_sizes))
    ]

    return [Directory(entries=dir_entries[:size]) for size in dir_sizes]


def test_write_large_objects(
    kafka_prefix: str,
    kafka_server: Tuple[Popen, int],
    consumer: Consumer,
    large_directories: List[Directory],
):
    kafka_prefix += ".swh.journal.objects"

    # Needed as there is no directories in TEST_OBJECT_DICTS, the consumer
    # isn't autosubscribed to directories.
    consumer.subscribe([kafka_prefix + ".directory"])

    writer = KafkaJournalWriter(
        brokers=["localhost:%d" % kafka_server[1]],
        client_id="kafka_writer",
        prefix=kafka_prefix,
    )

    writer.write_additions("directory", large_directories)

    consumed_messages = consume_messages(consumer, kafka_prefix, len(large_directories))

    for dir, message in zip(large_directories, consumed_messages["directory"]):
        (dir_id, consumed_dir) = message
        assert dir_id == dir.id
        assert consumed_dir == dir.to_dict()


def dir_message_size(directory: Directory) -> int:
    """Estimate the size of a directory kafka message.

    We could just do it with `len(value_to_kafka(directory.to_dict()))`,
    but the serialization is a substantial chunk of the test time here.

    """
    n_entries = len(directory.entries)
    return (
        # fmt: off
        0
        + 1       # header of a 2-element fixmap
        + 1 + 2   # fixstr("id")
        + 2 + 20  # bin8(directory.id of length 20)
        + 1 + 7   # fixstr("entries")
        + 4       # array header
        + n_entries
        * (
            0
            + 1       # header of a 4-element fixmap
            + 1 + 6   # fixstr("target")
            + 2 + 20  # bin8(target of length 20)
            + 1 + 4   # fixstr("name")
            + 2 + 9   # bin8(name of length 9)
            + 1 + 5   # fixstr("perms")
            + 5       # uint32(perms)
            + 1 + 4   # fixstr("type")
            + 1 + 3   # fixstr(type)
        )
        # fmt: on
    )


SMALL_MESSAGE_SIZE = 1024 * 1024


@pytest.mark.parametrize(
    "kafka_server_config_overrides", [{"message.max.bytes": str(SMALL_MESSAGE_SIZE)}]
)
def test_fail_write_large_objects(
    kafka_prefix: str,
    kafka_server: Tuple[Popen, int],
    consumer: Consumer,
    large_directories: List[Directory],
):
    kafka_prefix += ".swh.journal.objects"

    # Needed as there is no directories in TEST_OBJECT_DICTS, the consumer
    # isn't autosubscribed to directories.
    consumer.subscribe([kafka_prefix + ".directory"])

    writer = KafkaJournalWriter(
        brokers=["localhost:%d" % kafka_server[1]],
        client_id="kafka_writer",
        prefix=kafka_prefix,
    )

    expected_dirs = []

    for directory in large_directories:
        if dir_message_size(directory) < SMALL_MESSAGE_SIZE:
            # No error; write anyway, but continue
            writer.write_addition("directory", directory)
            expected_dirs.append(directory)
            continue

        with pytest.raises(KafkaDeliveryError) as exc:
            writer.write_addition("directory", directory)

        assert "Failed deliveries" in exc.value.message
        assert len(exc.value.delivery_failures) == 1

        object_type, key, msg, code = exc.value.delivery_failures[0]

        assert object_type == "directory"
        assert key == directory.id
        assert code == "MSG_SIZE_TOO_LARGE"

    consumed_messages = consume_messages(consumer, kafka_prefix, len(expected_dirs))

    for dir, message in zip(expected_dirs, consumed_messages["directory"]):
        (dir_id, consumed_dir) = message
        assert dir_id == dir.id
        assert consumed_dir == dir.to_dict()


def test_write_delivery_timeout(
    kafka_prefix: str, kafka_server: Tuple[Popen, int], consumer: Consumer
):

    produced = []

    class MockProducer(Producer):
        def produce(self, **kwargs):
            produced.append(kwargs)

    kafka_prefix += ".swh.journal.objects"
    writer = KafkaJournalWriter(
        brokers=["localhost:%d" % kafka_server[1]],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        flush_timeout=1,
        producer_class=MockProducer,
    )

    empty_dir = Directory(entries=[])
    with pytest.raises(KafkaDeliveryError) as exc:
        writer.write_addition("directory", empty_dir)

    assert len(produced) == 1

    assert "timeout" in exc.value.message
    assert len(exc.value.delivery_failures) == 1
    delivery_failure = exc.value.delivery_failures[0]
    assert delivery_failure.key == empty_dir.id
    assert delivery_failure.code == "SWH_FLUSH_TIMEOUT"



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter
import copy
import functools
import logging
import re
import tempfile
from subprocess import Popen
from typing import Any, Dict, Tuple
from unittest.mock import patch

from click.testing import CliRunner
from confluent_kafka import Producer
import pytest
import yaml

from swh.model.hashutil import hash_to_hex
from swh.objstorage.backends.in_memory import InMemoryObjStorage
from swh.storage import get_storage

from swh.journal.cli import cli
from swh.journal.replay import CONTENT_REPLAY_RETRIES
from swh.journal.serializers import key_to_kafka, value_to_kafka


logger = logging.getLogger(__name__)


CLI_CONFIG = {
    "storage": {"cls": "memory",},
    "objstorage_src": {"cls": "mocked", "name": "src",},
    "objstorage_dst": {"cls": "mocked", "name": "dst",},
}


@pytest.fixture
def storage():
    """An swh-storage object that gets injected into the CLI functions."""
    storage_config = {"cls": "pipeline", "steps": [{"cls": "memory"},]}
    storage = get_storage(**storage_config)
    with patch("swh.journal.cli.get_storage") as get_storage_mock:
        get_storage_mock.return_value = storage
        yield storage


@pytest.fixture
def monkeypatch_retry_sleep(monkeypatch):
    from swh.journal.replay import copy_object, obj_in_objstorage

    monkeypatch.setattr(copy_object.retry, "sleep", lambda x: None)
    monkeypatch.setattr(obj_in_objstorage.retry, "sleep", lambda x: None)


def invoke(*args, env=None, journal_config=None):
    config = copy.deepcopy(CLI_CONFIG)
    if journal_config:
        config["journal"] = journal_config

    runner = CliRunner()
    with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
        yaml.dump(config, config_fd)
        config_fd.seek(0)
        args = ["-C" + config_fd.name] + list(args)
        return runner.invoke(cli, args, obj={"log_level": logging.DEBUG}, env=env,)


def test_replay(
    storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
):
    (_, kafka_port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    producer = Producer(
        {
            "bootstrap.servers": "localhost:{}".format(kafka_port),
            "client.id": "test-producer",
            "acks": "all",
        }
    )

    snapshot = {
        "id": b"foo",
        "branches": {b"HEAD": {"target_type": "revision", "target": b"\x01" * 20,}},
    }  # type: Dict[str, Any]
    producer.produce(
        topic=kafka_prefix + ".snapshot",
        key=key_to_kafka(snapshot["id"]),
        value=value_to_kafka(snapshot),
    )
    producer.flush()

    logger.debug("Flushed producer")

    result = invoke(
        "replay",
        "--stop-after-objects",
        "1",
        journal_config={
            "brokers": ["127.0.0.1:%d" % kafka_port],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
        },
    )

    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    assert storage.snapshot_get(snapshot["id"]) == {**snapshot, "next_branch": None}


def _patch_objstorages(names):
    objstorages = {name: InMemoryObjStorage() for name in names}

    def get_mock_objstorage(cls, **args):
        assert cls == "mocked", cls
        return objstorages[args["name"]]

    def decorator(f):
        @functools.wraps(f)
        @patch("swh.journal.cli.get_objstorage")
        def newf(get_objstorage_mock, *args, **kwargs):
            get_objstorage_mock.side_effect = get_mock_objstorage
            f(*args, objstorages=objstorages, **kwargs)

        return newf

    return decorator


NUM_CONTENTS = 10


def _fill_objstorage_and_kafka(kafka_port, kafka_prefix, objstorages):
    producer = Producer(
        {
            "bootstrap.servers": "127.0.0.1:{}".format(kafka_port),
            "client.id": "test-producer",
            "acks": "all",
        }
    )

    contents = {}
    for i in range(NUM_CONTENTS):
        content = b"\x00" * 19 + bytes([i])
        sha1 = objstorages["src"].add(content)
        contents[sha1] = content
        producer.produce(
            topic=kafka_prefix + ".content",
            key=key_to_kafka(sha1),
            value=key_to_kafka({"sha1": sha1, "status": "visible",}),
        )

    producer.flush()

    return contents


@_patch_objstorages(["src", "dst"])
def test_replay_content(
    objstorages,
    storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
):
    (_, kafka_port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    contents = _fill_objstorage_and_kafka(kafka_port, kafka_prefix, objstorages)

    result = invoke(
        "content-replay",
        "--stop-after-objects",
        str(NUM_CONTENTS),
        journal_config={
            "brokers": ["127.0.0.1:%d" % kafka_port],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
        },
    )

    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    for (sha1, content) in contents.items():
        assert sha1 in objstorages["dst"], sha1
        assert objstorages["dst"].get(sha1) == content


@_patch_objstorages(["src", "dst"])
def test_replay_content_structured_log(
    objstorages,
    storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
    caplog,
):
    (_, kafka_port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    contents = _fill_objstorage_and_kafka(kafka_port, kafka_prefix, objstorages)

    caplog.set_level(logging.DEBUG, "swh.journal.replay")

    expected_obj_ids = set(hash_to_hex(sha1) for sha1 in contents)

    result = invoke(
        "content-replay",
        "--stop-after-objects",
        str(NUM_CONTENTS),
        journal_config={
            "brokers": ["127.0.0.1:%d" % kafka_port],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
        },
    )
    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    copied = set()
    for record in caplog.records:
        logtext = record.getMessage()
        if "copied" in logtext:
            copied.add(record.args["obj_id"])

    assert (
        copied == expected_obj_ids
    ), "Mismatched logging; see captured log output for details."


@_patch_objstorages(["src", "dst"])
def test_replay_content_static_group_id(
    objstorages,
    storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
    caplog,
):
    (_, kafka_port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    contents = _fill_objstorage_and_kafka(kafka_port, kafka_prefix, objstorages)

    # Setup log capture to fish the consumer settings out of the log messages
    caplog.set_level(logging.DEBUG, "swh.journal.client")

    result = invoke(
        "content-replay",
        "--stop-after-objects",
        str(NUM_CONTENTS),
        env={"KAFKA_GROUP_INSTANCE_ID": "static-group-instance-id"},
        journal_config={
            "brokers": ["127.0.0.1:%d" % kafka_port],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
        },
    )
    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    consumer_settings = None
    for record in caplog.records:
        if "Consumer settings" in record.message:
            consumer_settings = record.args
            break

    assert consumer_settings is not None, (
        "Failed to get consumer settings out of the consumer log. "
        "See log capture for details."
    )
    assert consumer_settings["group.instance.id"] == "static-group-instance-id"
    assert consumer_settings["session.timeout.ms"] == 60 * 10 * 1000
    assert consumer_settings["max.poll.interval.ms"] == 90 * 10 * 1000

    for (sha1, content) in contents.items():
        assert sha1 in objstorages["dst"], sha1
        assert objstorages["dst"].get(sha1) == content


@_patch_objstorages(["src", "dst"])
def test_replay_content_exclude(
    objstorages,
    storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
):
    (_, kafka_port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    contents = _fill_objstorage_and_kafka(kafka_port, kafka_prefix, objstorages)

    excluded_contents = list(contents)[0::2]  # picking half of them
    with tempfile.NamedTemporaryFile(mode="w+b") as fd:
        fd.write(b"".join(sorted(excluded_contents)))

        fd.seek(0)

        result = invoke(
            "content-replay",
            "--stop-after-objects",
            str(NUM_CONTENTS),
            "--exclude-sha1-file",
            fd.name,
            journal_config={
                "brokers": ["127.0.0.1:%d" % kafka_port],
                "group_id": kafka_consumer_group,
                "prefix": kafka_prefix,
            },
        )
    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    for (sha1, content) in contents.items():
        if sha1 in excluded_contents:
            assert sha1 not in objstorages["dst"], sha1
        else:
            assert sha1 in objstorages["dst"], sha1
            assert objstorages["dst"].get(sha1) == content


NUM_CONTENTS_DST = 5


@_patch_objstorages(["src", "dst"])
@pytest.mark.parametrize(
    "check_dst,expected_copied,expected_in_dst",
    [
        (True, NUM_CONTENTS - NUM_CONTENTS_DST, NUM_CONTENTS_DST),
        (False, NUM_CONTENTS, 0),
    ],
)
def test_replay_content_check_dst(
    objstorages,
    storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
    check_dst: bool,
    expected_copied: int,
    expected_in_dst: int,
    caplog,
):
    (_, kafka_port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    contents = _fill_objstorage_and_kafka(kafka_port, kafka_prefix, objstorages)

    for i, (sha1, content) in enumerate(contents.items()):
        if i >= NUM_CONTENTS_DST:
            break

        objstorages["dst"].add(content, obj_id=sha1)

    caplog.set_level(logging.DEBUG, "swh.journal.replay")

    result = invoke(
        "content-replay",
        "--stop-after-objects",
        str(NUM_CONTENTS),
        "--check-dst" if check_dst else "--no-check-dst",
        journal_config={
            "brokers": ["127.0.0.1:%d" % kafka_port],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
        },
    )
    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    copied = 0
    in_dst = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if "copied" in logtext:
            copied += 1
        elif "in dst" in logtext:
            in_dst += 1

    assert (
        copied == expected_copied and in_dst == expected_in_dst
    ), "Unexpected amount of objects copied, see the captured log for details"

    for (sha1, content) in contents.items():
        assert sha1 in objstorages["dst"], sha1
        assert objstorages["dst"].get(sha1) == content


class FlakyObjStorage(InMemoryObjStorage):
    def __init__(self, *args, **kwargs):
        state = kwargs.pop("state")
        self.failures_left = Counter(kwargs.pop("failures"))
        super().__init__(*args, **kwargs)
        if state:
            self.state = state

    def flaky_operation(self, op, obj_id):
        if self.failures_left[op, obj_id] > 0:
            self.failures_left[op, obj_id] -= 1
            raise RuntimeError("Failed %s on %s" % (op, hash_to_hex(obj_id)))

    def get(self, obj_id):
        self.flaky_operation("get", obj_id)
        return super().get(obj_id)

    def add(self, data, obj_id=None, check_presence=True):
        self.flaky_operation("add", obj_id)
        return super().add(data, obj_id=obj_id, check_presence=check_presence)

    def __contains__(self, obj_id):
        self.flaky_operation("in", obj_id)
        return super().__contains__(obj_id)


@_patch_objstorages(["src", "dst"])
def test_replay_content_check_dst_retry(
    objstorages,
    storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
    monkeypatch_retry_sleep,
):
    (_, kafka_port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    contents = _fill_objstorage_and_kafka(kafka_port, kafka_prefix, objstorages)

    failures = {}
    for i, (sha1, content) in enumerate(contents.items()):
        if i >= NUM_CONTENTS_DST:
            break

        objstorages["dst"].add(content, obj_id=sha1)
        failures["in", sha1] = 1

    orig_dst = objstorages["dst"]
    objstorages["dst"] = FlakyObjStorage(state=orig_dst.state, failures=failures)

    result = invoke(
        "content-replay",
        "--stop-after-objects",
        str(NUM_CONTENTS),
        "--check-dst",
        journal_config={
            "brokers": ["127.0.0.1:%d" % kafka_port],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
        },
    )
    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    for (sha1, content) in contents.items():
        assert sha1 in objstorages["dst"], sha1
        assert objstorages["dst"].get(sha1) == content


@_patch_objstorages(["src", "dst"])
def test_replay_content_failed_copy_retry(
    objstorages,
    storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
    caplog,
    monkeypatch_retry_sleep,
):
    (_, kafka_port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    contents = _fill_objstorage_and_kafka(kafka_port, kafka_prefix, objstorages)

    add_failures = {}
    get_failures = {}
    definitely_failed = set()

    # We want to generate operations failing 1 to CONTENT_REPLAY_RETRIES times.
    # We generate failures for 2 different operations, get and add.
    num_retry_contents = 2 * CONTENT_REPLAY_RETRIES

    assert (
        num_retry_contents < NUM_CONTENTS
    ), "Need to generate more test contents to properly test retry behavior"

    for i, sha1 in enumerate(contents):
        if i >= num_retry_contents:
            break

        # This generates a number of failures, up to CONTENT_REPLAY_RETRIES
        num_failures = (i % CONTENT_REPLAY_RETRIES) + 1

        # This generates failures of add for the first CONTENT_REPLAY_RETRIES
        # objects, then failures of get.
        if i < CONTENT_REPLAY_RETRIES:
            add_failures["add", sha1] = num_failures
        else:
            get_failures["get", sha1] = num_failures

        # Only contents that have CONTENT_REPLAY_RETRIES or more are
        # definitely failing
        if num_failures >= CONTENT_REPLAY_RETRIES:
            definitely_failed.add(hash_to_hex(sha1))

    objstorages["dst"] = FlakyObjStorage(
        state=objstorages["dst"].state, failures=add_failures,
    )
    objstorages["src"] = FlakyObjStorage(
        state=objstorages["src"].state, failures=get_failures,
    )

    caplog.set_level(logging.DEBUG, "swh.journal.replay")

    result = invoke(
        "content-replay",
        "--stop-after-objects",
        str(NUM_CONTENTS),
        journal_config={
            "brokers": ["127.0.0.1:%d" % kafka_port],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
        },
    )
    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    copied = 0
    actually_failed = set()
    for record in caplog.records:
        logtext = record.getMessage()
        if "copied" in logtext:
            copied += 1
        elif "Failed operation" in logtext:
            assert record.levelno == logging.ERROR
            assert record.args["retries"] == CONTENT_REPLAY_RETRIES
            actually_failed.add(record.args["obj_id"])

    assert (
        actually_failed == definitely_failed
    ), "Unexpected object copy failures; see captured log for details"

    for (sha1, content) in contents.items():
        if hash_to_hex(sha1) in definitely_failed:
            assert sha1 not in objstorages["dst"]
            continue

        assert sha1 in objstorages["dst"], sha1
        assert objstorages["dst"].get(sha1) == content


@_patch_objstorages(["src", "dst"])
def test_replay_content_objnotfound(
    objstorages,
    storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
    caplog,
):
    (_, kafka_port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    contents = _fill_objstorage_and_kafka(kafka_port, kafka_prefix, objstorages)

    num_contents_deleted = 5
    contents_deleted = set()

    for i, sha1 in enumerate(contents):
        if i >= num_contents_deleted:
            break

        del objstorages["src"].state[sha1]
        contents_deleted.add(hash_to_hex(sha1))

    caplog.set_level(logging.DEBUG, "swh.journal.replay")

    result = invoke(
        "content-replay",
        "--stop-after-objects",
        str(NUM_CONTENTS),
        journal_config={
            "brokers": ["127.0.0.1:%d" % kafka_port],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
        },
    )
    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    copied = 0
    not_in_src = set()
    for record in caplog.records:
        logtext = record.getMessage()
        if "copied" in logtext:
            copied += 1
        elif "object not found" in logtext:
            # Check that the object id can be recovered from logs
            assert record.levelno == logging.ERROR
            not_in_src.add(record.args["obj_id"])

    assert (
        copied == NUM_CONTENTS - num_contents_deleted
    ), "Unexpected number of contents copied"

    assert (
        not_in_src == contents_deleted
    ), "Mismatch between deleted contents and not_in_src logs"

    for (sha1, content) in contents.items():
        if sha1 not in objstorages["src"]:
            continue
        assert sha1 in objstorages["dst"], sha1
        assert objstorages["dst"].get(sha1) == content



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from subprocess import Popen
from typing import Dict, List, Tuple
from unittest.mock import MagicMock

from confluent_kafka import Producer
import pytest

from swh.model.hypothesis_strategies import revisions
from swh.model.model import Content

from swh.journal.client import JournalClient
from swh.journal.serializers import key_to_kafka, value_to_kafka


def test_client(
    kafka_prefix: str, kafka_consumer_group: str, kafka_server: Tuple[Popen, int]
):
    (_, port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    producer = Producer(
        {
            "bootstrap.servers": "localhost:{}".format(port),
            "client.id": "test producer",
            "acks": "all",
        }
    )

    rev = revisions().example()

    # Fill Kafka
    producer.produce(
        topic=kafka_prefix + ".revision",
        key=key_to_kafka(rev.id),
        value=value_to_kafka(rev.to_dict()),
    )
    producer.flush()

    client = JournalClient(
        brokers="localhost:%d" % kafka_server[1],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_after_objects=1,
    )
    worker_fn = MagicMock()
    client.process(worker_fn)

    worker_fn.assert_called_once_with({"revision": [rev.to_dict()]})


def test_client_eof(
    kafka_prefix: str, kafka_consumer_group: str, kafka_server: Tuple[Popen, int]
):
    (_, port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    producer = Producer(
        {
            "bootstrap.servers": "localhost:{}".format(port),
            "client.id": "test producer",
            "acks": "all",
        }
    )

    rev = revisions().example()

    # Fill Kafka
    producer.produce(
        topic=kafka_prefix + ".revision",
        key=key_to_kafka(rev.id),
        value=value_to_kafka(rev.to_dict()),
    )
    producer.flush()

    client = JournalClient(
        brokers="localhost:%d" % kafka_server[1],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_after_objects=None,
        stop_on_eof=True,
    )

    worker_fn = MagicMock()
    client.process(worker_fn)

    worker_fn.assert_called_once_with({"revision": [rev.to_dict()]})


@pytest.mark.parametrize("batch_size", [1, 5, 100])
def test_client_batch_size(
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: Tuple[Popen, int],
    batch_size: int,
):
    (_, port) = kafka_server
    kafka_prefix += ".swh.journal.objects"

    num_objects = 2 * batch_size + 1
    assert num_objects < 256, "Too many objects, generation will fail"

    producer = Producer(
        {
            "bootstrap.servers": "localhost:{}".format(port),
            "client.id": "test producer",
            "acks": "all",
        }
    )

    contents = [Content.from_data(bytes([i])) for i in range(num_objects)]

    # Fill Kafka
    for content in contents:
        producer.produce(
            topic=kafka_prefix + ".content",
            key=key_to_kafka(content.sha1),
            value=value_to_kafka(content.to_dict()),
        )

    producer.flush()

    client = JournalClient(
        brokers=["localhost:%d" % kafka_server[1]],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_after_objects=num_objects,
        batch_size=batch_size,
    )

    collected_output: List[Dict] = []

    def worker_fn(objects):
        received = objects["content"]
        assert len(received) <= batch_size
        collected_output.extend(received)

    client.process(worker_fn)

    assert collected_output == [content.to_dict() for content in contents]



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
from unittest.mock import patch

import attr
from hypothesis import given, settings, HealthCheck
from hypothesis.strategies import lists

from swh.model.hypothesis_strategies import present_contents
from swh.model.model import Origin
from swh.storage import get_storage
from swh.storage.exc import HashCollision

from swh.journal.replay import (
    process_replay_objects,
    process_replay_objects_content,
    object_converter_fn,
)

from .utils import MockedJournalClient, MockedKafkaWriter
from .conftest import objects_d

storage_config = {
    "cls": "memory",
    "journal_writer": {"cls": "memory"},
}


def empty_person_name_email(rev_or_rel):
    """Empties the 'name' and 'email' fields of the author/committer fields
    of a revision or release; leaving only the fullname."""
    if getattr(rev_or_rel, "author", None):
        rev_or_rel = attr.evolve(
            rev_or_rel, author=attr.evolve(rev_or_rel.author, name=b"", email=b"",)
        )

    if getattr(rev_or_rel, "committer", None):
        rev_or_rel = attr.evolve(
            rev_or_rel,
            committer=attr.evolve(rev_or_rel.committer, name=b"", email=b"",),
        )

    return rev_or_rel


@given(lists(objects_d(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_write_replay_same_order_batches(objects):
    queue = []
    replayer = MockedJournalClient(queue)

    with patch(
        "swh.journal.writer.inmemory.InMemoryJournalWriter",
        return_value=MockedKafkaWriter(queue),
    ):
        storage1 = get_storage(**storage_config)

    # Write objects to storage1
    for (obj_type, obj) in objects:
        if obj_type == "content" and obj.get("status") == "absent":
            obj_type = "skipped_content"

        obj = object_converter_fn[obj_type](obj)

        if obj_type == "origin_visit":
            storage1.origin_add_one(Origin(url=obj.origin))
            storage1.origin_visit_upsert([obj])
        else:
            method = getattr(storage1, obj_type + "_add")
            try:
                method([obj])
            except HashCollision:
                pass

    # Bail out early if we didn't insert any relevant objects...
    queue_size = len(queue)
    assert queue_size != 0, "No test objects found; hypothesis strategy bug?"

    assert replayer.stop_after_objects is None
    replayer.stop_after_objects = queue_size

    storage2 = get_storage(**storage_config)
    worker_fn = functools.partial(process_replay_objects, storage=storage2)

    replayer.process(worker_fn)

    assert replayer.consumer.committed

    for attr_name in (
        "_contents",
        "_directories",
        "_snapshots",
        "_origin_visits",
        "_origins",
    ):
        assert getattr(storage1, attr_name) == getattr(storage2, attr_name), attr_name

    # When hypothesis generates a revision and a release with same
    # author (or committer) fullname but different name or email, then
    # the storage will use the first name/email it sees.
    # This first one will be either the one from the revision or the release,
    # and since there is no order guarantees, storage2 has 1/2 chance of
    # not seeing the same order as storage1, therefore we need to strip
    # them out before comparing.
    for attr_name in ("_revisions", "_releases"):
        items1 = {
            k: empty_person_name_email(v)
            for (k, v) in getattr(storage1, attr_name).items()
        }
        items2 = {
            k: empty_person_name_email(v)
            for (k, v) in getattr(storage2, attr_name).items()
        }
        assert items1 == items2, attr_name


# TODO: add test for hash collision


@given(lists(present_contents(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_write_replay_content(objects):

    queue = []
    replayer = MockedJournalClient(queue)

    with patch(
        "swh.journal.writer.inmemory.InMemoryJournalWriter",
        return_value=MockedKafkaWriter(queue),
    ):
        storage1 = get_storage(**storage_config)

    contents = []
    for obj in objects:
        storage1.content_add([obj])
        contents.append(obj)

    # Bail out early if we didn't insert any relevant objects...
    queue_size = len(queue)
    assert queue_size != 0, "No test objects found; hypothesis strategy bug?"

    assert replayer.stop_after_objects is None
    replayer.stop_after_objects = queue_size

    storage2 = get_storage(**storage_config)

    objstorage1 = storage1.objstorage.objstorage
    objstorage2 = storage2.objstorage.objstorage

    worker_fn = functools.partial(
        process_replay_objects_content, src=objstorage1, dst=objstorage2
    )

    replayer.process(worker_fn)

    # only content with status visible will be copied in storage2
    expected_objstorage_state = {
        c.sha1: c.data for c in contents if c.status == "visible"
    }

    assert expected_objstorage_state == objstorage2.state



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import os
import pytest
import logging
import random
import string

from confluent_kafka import Consumer
from confluent_kafka.admin import AdminClient, ConfigResource

from hypothesis.strategies import one_of
from subprocess import Popen
from typing import Any, Dict, Iterator, List, Tuple

from pathlib import Path
from pytest_kafka import (
    make_zookeeper_process,
    make_kafka_server,
    KAFKA_SERVER_CONFIG_TEMPLATE,
    ZOOKEEPER_CONFIG_TEMPLATE,
)

from swh.model import hypothesis_strategies as strategies
from swh.model.hashutil import MultiHash, hash_to_bytes

from swh.journal.serializers import ModelObject
from swh.journal.writer.kafka import OBJECT_TYPES


logger = logging.getLogger(__name__)

CONTENTS = [
    {**MultiHash.from_data(b"foo").digest(), "length": 3, "status": "visible",},
]

duplicate_content1 = {
    "length": 4,
    "sha1": hash_to_bytes("44973274ccef6ab4dfaaf86599792fa9c3fe4689"),
    "sha1_git": b"another-foo",
    "blake2s256": b"another-bar",
    "sha256": b"another-baz",
    "status": "visible",
}

# Craft a sha1 collision
duplicate_content2 = duplicate_content1.copy()
sha1_array = bytearray(duplicate_content1["sha1_git"])
sha1_array[0] += 1
duplicate_content2["sha1_git"] = bytes(sha1_array)


DUPLICATE_CONTENTS = [duplicate_content1, duplicate_content2]


COMMITTERS = [
    {"fullname": b"foo", "name": b"foo", "email": b"",},
    {"fullname": b"bar", "name": b"bar", "email": b"",},
]

DATES = [
    {
        "timestamp": {"seconds": 1234567891, "microseconds": 0,},
        "offset": 120,
        "negative_utc": False,
    },
    {
        "timestamp": {"seconds": 1234567892, "microseconds": 0,},
        "offset": 120,
        "negative_utc": False,
    },
]

REVISIONS = [
    {
        "id": hash_to_bytes("7026b7c1a2af56521e951c01ed20f255fa054238"),
        "message": b"hello",
        "date": DATES[0],
        "committer": COMMITTERS[0],
        "author": COMMITTERS[0],
        "committer_date": DATES[0],
        "type": "git",
        "directory": b"\x01" * 20,
        "synthetic": False,
        "metadata": None,
        "parents": [],
    },
    {
        "id": hash_to_bytes("368a48fe15b7db2383775f97c6b247011b3f14f4"),
        "message": b"hello again",
        "date": DATES[1],
        "committer": COMMITTERS[1],
        "author": COMMITTERS[1],
        "committer_date": DATES[1],
        "type": "hg",
        "directory": b"\x02" * 20,
        "synthetic": False,
        "metadata": None,
        "parents": [],
    },
]

RELEASES = [
    {
        "id": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
        "name": b"v0.0.1",
        "date": {
            "timestamp": {"seconds": 1234567890, "microseconds": 0,},
            "offset": 120,
            "negative_utc": False,
        },
        "author": COMMITTERS[0],
        "target_type": "revision",
        "target": b"\x04" * 20,
        "message": b"foo",
        "synthetic": False,
    },
]

ORIGINS = [
    {"url": "https://somewhere.org/den/fox",},
    {"url": "https://overtherainbow.org/fox/den",},
]

ORIGIN_VISITS = [
    {
        "origin": ORIGINS[0]["url"],
        "date": "2013-05-07 04:20:39.369271+00:00",
        "snapshot": None,  # TODO
        "status": "ongoing",  # TODO
        "metadata": {"foo": "bar"},
        "type": "git",
    },
    {
        "origin": ORIGINS[0]["url"],
        "date": "2018-11-27 17:20:39+00:00",
        "snapshot": None,  # TODO
        "status": "ongoing",  # TODO
        "metadata": {"baz": "qux"},
        "type": "git",
    },
]

TEST_OBJECT_DICTS: Dict[str, List[Dict[str, Any]]] = {
    "content": CONTENTS,
    "revision": REVISIONS,
    "release": RELEASES,
    "origin": ORIGINS,
    "origin_visit": ORIGIN_VISITS,
}

MODEL_OBJECTS = {v: k for (k, v) in OBJECT_TYPES.items()}

TEST_OBJECTS: Dict[str, List[ModelObject]] = {}

for object_type, objects in TEST_OBJECT_DICTS.items():
    converted_objects: List[ModelObject] = []
    model = MODEL_OBJECTS[object_type]

    for (num, obj_d) in enumerate(objects):
        if object_type == "origin_visit":
            obj_d = {**obj_d, "visit": num}
        elif object_type == "content":
            obj_d = {**obj_d, "data": b"", "ctime": datetime.datetime.now()}

        converted_objects.append(model.from_dict(obj_d))

    TEST_OBJECTS[object_type] = converted_objects


KAFKA_ROOT = os.environ.get("SWH_KAFKA_ROOT")
KAFKA_ROOT = KAFKA_ROOT if KAFKA_ROOT else os.path.dirname(__file__) + "/kafka"
if not os.path.exists(KAFKA_ROOT):
    msg = (
        "Development error: %s must exist and target an "
        "existing kafka installation" % KAFKA_ROOT
    )
    raise ValueError(msg)

KAFKA_SCRIPTS = Path(KAFKA_ROOT) / "bin"

KAFKA_BIN = str(KAFKA_SCRIPTS / "kafka-server-start.sh")
ZOOKEEPER_BIN = str(KAFKA_SCRIPTS / "zookeeper-server-start.sh")

ZK_CONFIG_TEMPLATE = ZOOKEEPER_CONFIG_TEMPLATE + "\nadmin.enableServer=false\n"
KAFKA_CONFIG_TEMPLATE = KAFKA_SERVER_CONFIG_TEMPLATE + "\nmessage.max.bytes=104857600\n"

# Those defines fixtures
zookeeper_proc = make_zookeeper_process(
    ZOOKEEPER_BIN, zk_config_template=ZK_CONFIG_TEMPLATE, scope="session"
)
os.environ[
    "KAFKA_LOG4J_OPTS"
] = "-Dlog4j.configuration=file:%s/log4j.properties" % os.path.dirname(__file__)
session_kafka_server = make_kafka_server(
    KAFKA_BIN,
    "zookeeper_proc",
    kafka_config_template=KAFKA_CONFIG_TEMPLATE,
    scope="session",
)

kafka_logger = logging.getLogger("kafka")
kafka_logger.setLevel(logging.WARN)


@pytest.fixture(scope="function")
def kafka_prefix():
    """Pick a random prefix for kafka topics on each call"""
    return "".join(random.choice(string.ascii_lowercase) for _ in range(10))


@pytest.fixture(scope="function")
def kafka_consumer_group(kafka_prefix: str):
    """Pick a random consumer group for kafka consumers on each call"""
    return "test-consumer-%s" % kafka_prefix


@pytest.fixture(scope="session")
def kafka_admin_client(session_kafka_server: Tuple[Popen, int]) -> AdminClient:
    return AdminClient({"bootstrap.servers": "localhost:%s" % session_kafka_server[1]})


@pytest.fixture(scope="function")
def kafka_server_config_overrides() -> Dict[str, str]:
    return {}


@pytest.fixture(scope="function")
def kafka_server(
    session_kafka_server: Tuple[Popen, int],
    kafka_admin_client: AdminClient,
    kafka_server_config_overrides: Dict[str, str],
) -> Iterator[Tuple[Popen, int]]:
    # No overrides, we can just return the original broker connection
    if not kafka_server_config_overrides:
        yield session_kafka_server
        return

    # This is the minimal operation that the kafka_admin_client gives to
    # retrieve the cluster metadata, which we need to get the numeric id of the
    # broker spawned by pytest_kafka.
    metadata = kafka_admin_client.list_topics("__consumer_offsets")
    broker_ids = [str(broker) for broker in metadata.brokers.keys()]
    assert len(broker_ids) == 1, "More than one broker found in the kafka cluster?!"

    # Pull the current broker configuration. describe_configs and alter_configs
    # generate a dict containing one concurrent.future per queried
    # ConfigResource, hence the use of .result()
    broker = ConfigResource("broker", broker_ids[0])
    futures = kafka_admin_client.describe_configs([broker])
    original_config = futures[broker].result()

    # Gather the list of settings we need to change in the broker
    # ConfigResource, and their original values in the to_restore dict
    to_restore = {}
    for key, new_value in kafka_server_config_overrides.items():
        if key not in original_config:
            raise ValueError(f"Cannot override unknown configuration {key}")
        orig_value = original_config[key].value
        if orig_value == new_value:
            continue
        if original_config[key].is_read_only:
            raise ValueError(f"Cannot override read-only configuration {key}")

        broker.set_config(key, new_value)
        to_restore[key] = orig_value

    # to_restore will be empty if all the config "overrides" are equal to the
    # original value. No need to wait for a config alteration if that's the
    # case. The result() will raise a KafkaException if the settings change
    # failed.
    if to_restore:
        futures = kafka_admin_client.alter_configs([broker])
        try:
            futures[broker].result()
        except Exception:
            raise

    yield session_kafka_server

    # Now we can restore the old setting values. Again, the result() will raise
    # a KafkaException if the settings change failed.
    if to_restore:
        for key, orig_value in to_restore.items():
            broker.set_config(key, orig_value)

        futures = kafka_admin_client.alter_configs([broker])
        try:
            futures[broker].result()
        except Exception:
            raise


TEST_CONFIG = {
    "consumer_id": "swh.journal.consumer",
    "object_types": TEST_OBJECT_DICTS.keys(),
    "stop_after_objects": 1,  # will read 1 object and stop
    "storage": {"cls": "memory", "args": {}},
}


@pytest.fixture
def test_config(kafka_server: Tuple[Popen, int], kafka_prefix: str):
    """Test configuration needed for producer/consumer

    """
    _, port = kafka_server
    return {
        **TEST_CONFIG,
        "brokers": ["127.0.0.1:{}".format(port)],
        "prefix": kafka_prefix + ".swh.journal.objects",
    }


@pytest.fixture
def consumer(
    kafka_server: Tuple[Popen, int], test_config: Dict, kafka_consumer_group: str,
) -> Consumer:
    """Get a connected Kafka consumer.

    """
    _, kafka_port = kafka_server
    consumer = Consumer(
        {
            "bootstrap.servers": "127.0.0.1:{}".format(kafka_port),
            "auto.offset.reset": "earliest",
            "enable.auto.commit": True,
            "group.id": kafka_consumer_group,
        }
    )

    kafka_topics = [
        "%s.%s" % (test_config["prefix"], object_type)
        for object_type in test_config["object_types"]
    ]

    consumer.subscribe(kafka_topics)

    yield consumer

    consumer.close()


def objects_d():
    return one_of(
        strategies.origins().map(lambda x: ("origin", x.to_dict())),
        strategies.origin_visits().map(lambda x: ("origin_visit", x.to_dict())),
        strategies.snapshots().map(lambda x: ("snapshot", x.to_dict())),
        strategies.releases().map(lambda x: ("release", x.to_dict())),
        strategies.revisions().map(lambda x: ("revision", x.to_dict())),
        strategies.directories().map(lambda x: ("directory", x.to_dict())),
        strategies.skipped_contents().map(lambda x: ("skipped_content", x.to_dict())),
        strategies.present_contents().map(lambda x: ("content", x.to_dict())),
    )
